	/*
	 *		Copper Initialisation List
	 */

struct CopInit_AGA {
	copins vsync_hblank;
	copins fm0;
	copins bplcon0;
	copins diwstrt;
	copins diwstop;
	copins diwhigh;
	copins ddfstrt;
	copins ddfstop;
	copins bplcon2;
	copins sprfix[8];
	copins sprstrtup[16];
	copins wait;
	copins jump;
	copins wait_forever;
};


	/*
	 *		Long Frame/Short Frame Copper List
	 */

struct CopList_AGA {
	copins diwstrt;
	copins diwstop;
	copins diwhigh;
	copins ddfstrt;
	copins ddfstop;
	copins sprpt[2];		/* Sprite 0 */
	copins bplcon0;
	copins bplcon2;
	copins rest[64];
};


	/*
	 *		Hardware Cursor
	 */

struct CursorSprite_AGA {
	u_long Header[4];
	union {
		struct {
			u_long Data[64*4];
			u_long Trailer[4];
		} NonLaced;
		struct {
			u_long Data[32*4];
			u_long Trailer[4];
		} Laced;
	} u;
};

struct DummySprite_AGA {
	u_long Header[4];
	u_long Data[4];
	u_long Trailer[4];
};


	/*
	 *		Initialise the Copper Initialisation List
	 */

static void AGA_init_copinit(struct CopInit_AGA *copinit)
{
	int i, j;
	u_long p;

	copinit->vsync_hblank.l = CMOVE(0, CUSTOM_OFS(bplpt[0]));
	copinit->fm0.l = CMOVE(FMODE_SPAGEM|FMODE_SPR32|FMODE_BPAGEM|FMODE_BPL32,
								  CUSTOM_OFS(fmode));
	copinit->bplcon0.l = CMOVE(BPC0_COLOR|BPC0_SHRES|BPC0_ECSENA,
										CUSTOM_OFS(bplcon0));
	copinit->diwstrt.l = CMOVE(0x0181, CUSTOM_OFS(diwstrt));
	copinit->diwstop.l = CMOVE(0x0281, CUSTOM_OFS(diwstop));
	copinit->diwhigh.l = CMOVE(0x0000, CUSTOM_OFS(diwhigh));
	copinit->ddfstrt.l = CMOVE(0x0018, CUSTOM_OFS(ddfstrt));
	copinit->ddfstop.l = CMOVE(0x0020, CUSTOM_OFS(ddfstop));
	copinit->bplcon2.l = CMOVE(BPC2_KILLEHB|BPC2_PF2P2|BPC2_PF1P2,
										CUSTOM_OFS(bplcon2));
	for (i = 0; i < 8; i++)
		copinit->sprfix[i].l = CMOVE(0, CUSTOM_OFS(spr[i].pos));
	for (i = 0, j = 0; i < 8; i++) {
		p = (u_long)dummysprite;
		copinit->sprstrtup[j++].l = CMOVE(highw(p), CUSTOM_OFS(sprpt[i]));
		copinit->sprstrtup[j++].l = CMOVE(loww(p), CUSTOM_OFS(sprpt[i])+2);
	}
	copinit->wait.l = CWAIT(0, 10);
	copinit->jump.l = CMOVE(0, CUSTOM_OFS(copjmp2));
	copinit->wait_forever.l = CEND;

}


	/*
	 *		Rebuild the Copper Initialisation List
	 *
	 *		We only change the things that are not static
	 */

static void AGA_rebuild_copinit(struct amiga_fb_par *par,
										  struct CopInit_AGA *copinit)
{
	copinit->fm0.l = CMOVE(par->hw.AGA.fmode, CUSTOM_OFS(fmode));
	copinit->bplcon0.l = CMOVE(par->hw.AGA.bplcon0 & 
										~(BPC0_BPU3|BPC0_BPU2|BPC0_BPU1|BPC0_BPU0),
										CUSTOM_OFS(bplcon0));
	copinit->bplcon2.l = CMOVE(par->hw.AGA.bplcon2, CUSTOM_OFS(bplcon2));
}


	/*
	 *		Build the Copper List
	 */

static void AGA_build_clist(struct amiga_fb_par *par, struct CopList_AGA *cop,
									 struct CopList_AGA *othercop, u_short shf)
{
	u_short h_end2 = par->hw.AGA.htotal-4;
	u_short h_end3 = par->hw.AGA.htotal-11;
	u_long next_line = par->hw.AGA.next_line;
	u_long y_wrap, skip, p;
	int i, j = 0;

	cop->diwstrt.l = CMOVE(par->hw.AGA.diwstrt, CUSTOM_OFS(diwstrt));
	cop->diwstop.l = CMOVE(par->hw.AGA.diwstop, CUSTOM_OFS(diwstop));
	cop->diwhigh.l = CMOVE(par->hw.AGA.diwhigh, CUSTOM_OFS(diwhigh));
	cop->ddfstrt.l = CMOVE(par->hw.AGA.ddfstrt, CUSTOM_OFS(ddfstrt));
	cop->ddfstop.l = CMOVE(par->hw.AGA.ddfstop, CUSTOM_OFS(ddfstop));
	cop->bplcon0.l = CMOVE(par->hw.AGA.bplcon0, CUSTOM_OFS(bplcon0));
	cop->bplcon2.l = CMOVE(par->hw.AGA.bplcon2, CUSTOM_OFS(bplcon2));

	/* Point Sprite 0 at cursor sprite */

	if (shf) {
		p = par->hw.AGA.sprpt[1];
		cop->sprpt[0].l = CMOVE(highw(p), CUSTOM_OFS(sprpt[0]));
		cop->sprpt[1].l = CMOVE(loww(p), CUSTOM_OFS(sprpt[0])+2);
	} else {
		p = par->hw.AGA.sprpt[0];
		cop->sprpt[0].l = CMOVE(highw(p), CUSTOM_OFS(sprpt[0]));
		cop->sprpt[1].l = CMOVE(loww(p), CUSTOM_OFS(sprpt[0])+2);
	}

	skip = shf ? next_line : 0;
	if (par->var.vmode & FB_VMODE_WRAP) {
		y_wrap = par->var.yoffset;
		skip += y_wrap*next_line;
	} else
		y_wrap = 0;

	/* Set up initial bitplane ptrs */

	for (i = 0; i < par->var.bits_per_pixel; i++) {
		p = par->hw.AGA.bplpt[i]+skip;
		cop->rest[j++].l = CMOVE(highw(p), CUSTOM_OFS(bplpt[i]));
		cop->rest[j++].l = CMOVE(loww(p), CUSTOM_OFS(bplpt[i])+2);
	}

	if (y_wrap) {
		u_long line = par->hw.AGA.diwstrt_v;

		switch (par->var.vmode & FB_VMODE_MASK) {
			case FB_VMODE_NONINTERLACED:
				line += par->var.yres-y_wrap;
				break;
			case FB_VMODE_INTERLACED:
				line += (par->var.yres-y_wrap)/2;
				break;
			case FB_VMODE_DOUBLE:
				line += (par->var.yres-y_wrap)*2;
				break;
		}

		/* Handle skipping over 256-line chunks */

		while (line > 256) {
			/* Hardware limitation - 8 bit counter */
			cop->rest[j++].l = CWAIT(h_end2, 255);
			/* Wait(0, 0) - make sure we're in the new segment */
			cop->rest[j++].l = CWAIT(0, 0);
			line -= 256;
		}

		cop->rest[j++].l = CWAIT(h_end3, line-1);

		skip = shf ? next_line : 0;
		for (i = 0; i < par->var.bits_per_pixel; i++) {
			p = par->hw.AGA.bplpt[i]+skip;
			cop->rest[j++].l = CMOVE(highw(p), CUSTOM_OFS(bplpt[i]));
			cop->rest[j++].l = CMOVE(loww(p), CUSTOM_OFS(bplpt[i])+2);
		}
	}

	if (othercop) {
		cop->rest[j++].l = CMOVE(highw(othercop), CUSTOM_OFS(cop2lc));
		cop->rest[j++].l = CMOVE(loww(othercop), CUSTOM_OFS(cop2lc)+2);
	}

	/* End of Copper list */
	cop->rest[j++].l = CEND;
}


	/*
	 *		Initialisation
	 *
	 *		Allocate the required chip memory.
	 *		Set the default video mode for this chipset. If a video mode was
	 *		specified on the command line, it will override the default mode.
	 */

static int AGA_init(void)
{
	u_long chipptr;
	struct CopInit_AGA *copinit;

	/*
	 *		Disable Display DMA
	 */

	custom.dmacon = DMAF_ALL | DMAF_MASTER;

	/*
	 *		Set the Default Video Mode
	 */

	if (amifb_mode == -1)
		amifb_mode = AGA_DEFMODE;

	/*
	 *		Allocate Chip RAM Structures
	 */

	if (boot_info.bi_amiga.chip_size > 1048576)
		videomemorysize = VIDEOMEMSIZE_AGA_2M;
	else
		videomemorysize = VIDEOMEMSIZE_AGA_1M;

	chipptr = chipalloc(videomemorysize+
							  2*sizeof(struct CopInit_AGA)+
							  4*sizeof(struct CopList_AGA)+
							  2*sizeof(struct CursorSprite_AGA)+
							  sizeof(struct DummySprite_AGA));

	AssignChunk(videomemory, u_char *, chipptr, videomemorysize);
	AssignChunk(copdisplay[0].Init, copins *, chipptr,
					sizeof(struct CopInit_AGA));
	AssignChunk(copdisplay[0].LOFlist, copins *, chipptr,
					sizeof(struct CopList_AGA));
	AssignChunk(copdisplay[0].SHFlist, copins *, chipptr,
					sizeof(struct CopList_AGA));
	AssignChunk(copdisplay[1].Init, copins *, chipptr,
					sizeof(struct CopInit_AGA));
	AssignChunk(copdisplay[1].LOFlist, copins *, chipptr,
					sizeof(struct CopList_AGA));
	AssignChunk(copdisplay[1].SHFlist, copins *, chipptr,
					sizeof(struct CopList_AGA));
	AssignChunk(LOFsprite, u_long *, chipptr, sizeof(struct CursorSprite_AGA));
	AssignChunk(SHFsprite, u_long *, chipptr, sizeof(struct CursorSprite_AGA));
	AssignChunk(dummysprite, u_long *, chipptr, sizeof(struct DummySprite_AGA));

	/*
	 *		Make sure the Copper has something to do
	 */

	AGA_init_copinit((struct CopInit_AGA *)copdisplay[0].Init);
	AGA_init_copinit((struct CopInit_AGA *)copdisplay[1].Init);

	copinit = (struct CopInit_AGA *)copdisplay[0].Init;
	custom.cop1lc = (u_short *)copinit;
	custom.cop2lc = (u_short *)&copinit->wait_forever;
	custom.copjmp1 = 0;

	/*
	 *		Enable Display DMA
	 */

	custom.dmacon = DMAF_SETCLR|DMAF_MASTER|DMAF_RASTER|DMAF_COPPER|DMAF_SPRITE;

	return(0);
}


	/*
	 *		This function should fill in the `fix' structure based on the
	 *		values in the `par' structure.
	 */

static int AGA_encode_fix(struct fb_fix_screeninfo *fix,
								  struct amiga_fb_par *par)
{
	*fix = par->fix;
	return(0);
}


	/*
	 *		Get the video params out of `var'. If a value doesn't fit, round
	 *		it up, if it's too big, return EINVAL.
	 */

static int AGA_decode_var(struct fb_var_screeninfo *var,
								  struct amiga_fb_par *par)
{
	u_short diwstrt_v, diwstop_v, sprstrt_v, vsstrt, vsstop, vtotal;
	u_short diwstrt_h, diwstop_h, sprstrt_h, hsstrt, hsstop, htotal;
	u_short ddfmin, ddfmax, ddfstrt, ddfstop, hscroll, modulo;
	u_long bplpt, next_line, next_plane;
	int i;

	/*
	 *		Round up the Geometry Values (if necessary)
	 */

	if (var->xres == 0)
		var->xres = 64;
	if (var->yres == 0)
		var->yres = 64;
	if (var->xres > var->xres_virtual)
		var->xres_virtual = var->xres;
	if (var->yres > var->yres_virtual)
		var->yres_virtual = var->yres;
	var->xres_virtual = up64(var->xres_virtual);

	if (var->vmode & FB_VMODE_WRAP) {
		if (var->xoffset || (var->yoffset >= var->yres))
			return(-EINVAL);
	} else {
		if ((var->xoffset+var->xres > var->xres_virtual) ||
			 (var->yoffset+var->yres > var->yres_virtual))
			return(-EINVAL);
	}

	if (var->bits_per_pixel == 0)
		var->bits_per_pixel = 1;
	else if (var->bits_per_pixel > 8)
		return(-EINVAL);

	if (var->pixclock <= PixClock_SHRES)
		var->pixclock = PixClock_SHRES;
#if 0
	/*
	 *		TODO: Add support for HIRES and LORES later
	 */
	else if (var->pixclock <= PixClock_HIRES)
		var->pixclock = PixClock_HIRES;
	else if (var->pixclock <= PixClock_LORES)
		var->pixclock = PixClock_LORES;
#endif
	else
		return(-EINVAL);

	if (var->sync & FB_SYNC_BROADCAST) {
		/*
		 *		TODO: Add support for PAL/NTSC later
		 */
		return(-EINVAL);
	}

	/*
	 *		Vertical Timings
	 */

	switch (var->vmode & FB_VMODE_MASK) {
		case FB_VMODE_NONINTERLACED:
			vsstrt = var->lower_margin;
			vsstop = vsstrt+var->vsync_len;
			diwstrt_v = vsstop+var->upper_margin;
			diwstop_v = diwstrt_v+var->yres;
			vtotal = diwstop_v;
			sprstrt_v = diwstrt_v;
			break;

		case FB_VMODE_INTERLACED:
			if (var->yres & 1) {
				var->yres++;								/* even */
				if (var->yres > var->yres_virtual)
					var->yres_virtual = var->yres;
			}
			if (var->upper_margin & 1)
				var->upper_margin++;						/* even */
			if (!(var->lower_margin & 1))
				var->lower_margin++;						/* odd */
			if (var->vsync_len & 1)
				var->vsync_len++;							/* even */
			vsstrt = var->lower_margin/2;
			vsstop = vsstrt+var->vsync_len/2;
			diwstrt_v = vsstop+var->upper_margin/2;
			diwstop_v = diwstrt_v+var->yres/2;
			vtotal = diwstop_v;
			sprstrt_v = diwstrt_v;
			break;

		case FB_VMODE_DOUBLE:
			vsstrt = var->lower_margin*2;
			vsstop = vsstrt+var->vsync_len*2;
			diwstrt_v = vsstop+var->upper_margin*2;
			diwstop_v = diwstrt_v+var->yres*2;
			vtotal = diwstop_v;
			sprstrt_v = diwstrt_v;
			break;

		default:
			return(-EINVAL);
			break;
	}

	/*
	 *		Check the Memory Requirements
	 */

	if (var->xres_virtual*var->yres_virtual*var->bits_per_pixel/8 >
		 videomemorysize)
		return(-EINVAL);

	/*
	 *		Horizontal Timings
	 */

	var->hsync_len = up8(var->hsync_len);
	htotal = up8(var->left_margin+var->xres+var->right_margin+var->hsync_len);
	var->right_margin = htotal-var->left_margin-var->xres-var->hsync_len;
	hsstrt = down8(var->right_margin+4);
	hsstop = hsstrt+var->hsync_len;
	diwstop_h = htotal+hsstrt-var->right_margin+1;
	diwstrt_h = diwstop_h-var->xres-1;
	sprstrt_h = diwstrt_h-4;

	if (amifb_ilbm) {
		next_plane = div8(var->xres_virtual);
		next_line = var->bits_per_pixel*next_plane;
	} else {
		next_line = div8(var->xres_virtual);
		next_plane = var->yres_virtual*next_line;
	}

	/*
	 *		Display DMA Timings
	 */

	ddfmin = down64(var->xoffset);
	ddfmax = up64(var->xoffset+var->xres);
	ddfstrt = down64(diwstrt_h-68-mod64(var->xoffset));
	ddfstop = ddfstrt+ddfmax-ddfmin-64;
	bplpt = (u_long)videomemory+div8(ddfmin);

	if (!(var->vmode & FB_VMODE_WRAP))
		bplpt += var->yoffset*next_line;

	hscroll = diwstrt_h-68-ddfstrt-mod64(var->xoffset);
	modulo = next_line-div8(ddfmax-ddfmin);

	/*
	 *		Fill in the Variable Screeninfo
	 */

	var->grayscale = 0;
	var->red.offset = 0;
	var->red.length = 8;
	var->red.msb_right = 0;
	var->green.offset = 0;
	var->green.length = 8;
	var->green.msb_right = 0;
	var->blue.offset = 0;
	var->blue.length = 8;
	var->blue.msb_right = 0;
	var->transp.offset = 0;
	var->transp.length = 8;
	var->transp.msb_right = 0;
	var->nonstd = 0;
	var->height = -1;
	var->width = -1;
	var->accel = 0;

	par->var = *var;

	/*
	 *		Fill in the Fixed Screeninfo
	 */

	strcpy(par->fix.id, amiga_fb_name);
	par->fix.smem_start = (u_long)videomemory;
	par->fix.smem_len = videomemorysize;
	par->fix.reserved[0] = 0;
	par->fix.reserved[1] = 0;
	par->fix.reserved[2] = 0;
	par->fix.reserved[3] = 0;

	/*
	 *		Set the Type and Visual
	 */

	if (amifb_ilbm) {
		par->fix.type = FB_TYPE_INTERLEAVED_PLANES;
		par->fix.type_aux = next_line;
	} else {
		par->fix.type = FB_TYPE_PLANES;
		par->fix.type_aux = 0;
	}
	if (par->var.bits_per_pixel > 1)
		par->fix.visual = FB_VISUAL_PSEUDOCOLOR;
	else if (amifb_inverse)
		par->fix.visual = FB_VISUAL_MONO01;
	else
		par->fix.visual = FB_VISUAL_MONO10;

	/*
	 *		Can we do Smooth Panning?
	 */

	if (diwstrt_h >= 323)
		par->fix.xpanstep = 1;
	else
		par->fix.xpanstep = 64;
	par->fix.ypanstep = 1;

	/*
	 *		Can we do ywrap?
	 */

	/* TODO: This should consider the value of diwstrt_h too */

	if (par->var.bits_per_pixel <= 4)
		par->fix.ywrapstep = 1;
	else
		par->fix.ywrapstep = 0;

#if 0	/* This should not be in fb_fix_screeninfo */

	/*
	 *		Do we have a Hardware Cursor?
	 */

	if (diwstrt_h >= 260) {
		par->fix.crsr_width = 64;
		par->fix.crsr_height = 64;
		par->fix.crsr_relxres = 256;
		par->fix.crsr_relyres = 256;
		par->fix.crsr_color1 = 17;
		par->fix.crsr_color2 = 18;
	} else {
		par->fix.crsr_width = 0;
		par->fix.crsr_height = 0;
		par->fix.crsr_relxres = 0;
		par->fix.crsr_relyres = 0;
		par->fix.crsr_color1 = 0;
		par->fix.crsr_color2 = 0;
	}
#endif

	/*	
	 *		Calculate the Hardware Register Values
	 */

	par->hw.AGA.next_line = next_line;
	par->hw.AGA.next_plane = next_plane;
	par->hw.AGA.diwstrt_h = diwstrt_h;
	par->hw.AGA.diwstop_h = diwstop_h;
	par->hw.AGA.diwstrt_v = diwstrt_v;
	par->hw.AGA.diwstop_v = diwstop_v;
	par->hw.AGA.sprstrt_h = sprstrt_h;
	par->hw.AGA.sprstrt_v = sprstrt_v;

	par->hw.AGA.crsr_x = -1;
	par->hw.AGA.crsr_y = -1;

	for (i = 0; i < var->bits_per_pixel; i++)
		par->hw.AGA.bplpt[i] = bplpt+i*next_plane;
	par->hw.AGA.sprpt[0] = (u_long)LOFsprite;
	par->hw.AGA.sprpt[1] = (u_long)SHFsprite;
	par->hw.AGA.bplcon0 = BPC0_COLOR|BPC0_SHRES|BPC0_ECSENA;
	if (var->bits_per_pixel == 8)
		par->hw.AGA.bplcon0 |= BPC0_BPU3;
	else
		par->hw.AGA.bplcon0 |= var->bits_per_pixel<<12;
	if (var->sync & FB_SYNC_EXT)
		par->hw.AGA.bplcon0 |= BPC0_ERSY;
	par->hw.AGA.bplcon1 = (hscroll<<8 & 0xc000) | (hscroll<<12 & 0x3000) |
								 (hscroll<<4 & 0x0c00) | (hscroll<<8 & 0x0300) |
								 (hscroll<<2 & 0x00f0) | (hscroll>>2 & 0x000f);
	par->hw.AGA.bplcon2 = BPC2_KILLEHB|BPC2_PF2P2|BPC2_PF1P2;
	par->hw.AGA.bplcon3 = BPC3_PF2OF1|BPC3_PF2OF0|BPC3_SPRES1|BPC3_SPRES0|
								 BPC3_BRDRBLNK|BPC3_EXTBLKEN;
	par->hw.AGA.bplcon4 = BPC4_ESPRM4|BPC4_OSPRM4;
	par->hw.AGA.diwstrt = (diwstrt_v<<8 & 0xff00) | (diwstrt_h>>2 & 0x00ff);
	par->hw.AGA.diwstop = (diwstop_v<<8 & 0xff00) | (diwstop_h>>2 & 0x00ff);
	par->hw.AGA.diwhigh = (diwstop_h<<3 & 0x2000) | (diwstop_h<<11 & 0x1800) |
								 (diwstop_v & 0x0700) | (diwstrt_h>>5 & 0x0020) |
								 (diwstrt_h<<3 & 0x0018) | (diwstrt_v>>8 & 0x0007);
	par->hw.AGA.ddfstrt = div8(ddfstrt) & 0x00fe;
	par->hw.AGA.ddfstop = div8(ddfstop) & 0x00fe;
	par->hw.AGA.htotal = div8(htotal)-1;
	par->hw.AGA.hsstrt = div8(hsstrt);
	par->hw.AGA.hsstop = div8(hsstop);
	par->hw.AGA.hbstrt = div8(hsstrt);
	par->hw.AGA.hbstop = div8(hsstop);
	par->hw.AGA.vtotal = vtotal-1;
	par->hw.AGA.vsstrt = vsstrt;
	par->hw.AGA.vsstop = vsstop;
	par->hw.AGA.vbstrt = vsstrt;
	par->hw.AGA.vbstop = vsstop;
	par->hw.AGA.hcenter = div16(htotal)-1;
	par->hw.AGA.beamcon0 = BMC0_VARVBEN|BMC0_LOLDIS|BMC0_VARVSYEN|
								  BMC0_VARHSYEN|BMC0_VARBEAMEN|BMC0_PAL|BMC0_VARCSYEN;
	if (var->sync & FB_SYNC_HOR_HIGH_ACT)
		par->hw.AGA.beamcon0 |= BMC0_HSYTRUE;
	if (var->sync & FB_SYNC_VERT_HIGH_ACT)
		par->hw.AGA.beamcon0 |= BMC0_VSYTRUE;
	if (var->sync & FB_SYNC_COMP_HIGH_ACT)
		par->hw.AGA.beamcon0 |= BMC0_CSYTRUE;
	par->hw.AGA.fmode = FMODE_SPAGEM|FMODE_SPR32|FMODE_BPAGEM|FMODE_BPL32;

	switch (var->vmode & FB_VMODE_MASK) {
		case FB_VMODE_NONINTERLACED:
			par->hw.AGA.bpl1mod = modulo;
			par->hw.AGA.bpl2mod = modulo;
			break;

		case FB_VMODE_INTERLACED:
			par->hw.AGA.bpl1mod = modulo+next_line;
			par->hw.AGA.bpl2mod = modulo+next_line;
			par->hw.AGA.bplcon0 |= BPC0_LACE;
			break;

		case FB_VMODE_DOUBLE:
			par->hw.AGA.bpl1mod = modulo-next_line;
			par->hw.AGA.bpl2mod = modulo;
			par->hw.AGA.fmode |= FMODE_SSCAN2|FMODE_BSCAN2;
			break;
	}

	return(0);
}
	

	/*
	 *		Fill the `var' structure based on the values in `par' and maybe
	 *		other values read out of the hardware.
	 */

static int AGA_encode_var(struct fb_var_screeninfo *var,
								  struct amiga_fb_par *par)
{
	*var = par->var;
	return(0);
}


	/*
	 *		Fill the hardware's `par' structure.
	 */

static void AGA_get_par(struct amiga_fb_par *par)
{
	if (AGA_decode_var(&amiga_fb_predefined[amifb_mode], par))
		panic("Can't use default video mode\n");
}


	/*
	 *		Set the hardware according to `par'.
	 *
	 *		If `pan_only' is TRUE, don't change the whole video mode but only the
	 *		things required for a display pan or ywrap.
	 */

static void AGA_set_par(struct amiga_fb_par *par, u_short pan_only)
{
	struct CopDisplay *cd;

	do_vmode = 0;

	cd = &copdisplay[1-currentcop];
	if (pan_only) {

		/* TODO: Update the register data in current_par ... */

		/*
		 *		Panning may change the following registers:
		 *
		 *			bplcon1
		 *			bpl?mod
		 *			ddfstrt, ddfstop
		 *			bplpt*
		 */

	} else {

		/*
		 *		Set the Flag to indicate a Full Video Mode Change
		 *
		 *		This stays `1' until the Video Mode is changed during the VBlank
		 *		Interrupt. Note that subsequent calls of AGA_set_par(*, 1) do not
		 *		clear this flag, so the VBlank routine knows it still has to change
		 *		the Video Mode if that hasn't been done yet.
		 */

		full_vmode_change = 1;
	}
	AGA_rebuild_copinit(par, (struct CopInit_AGA *)cd->Init);
	if ((par->var.vmode & FB_VMODE_MASK) == FB_VMODE_INTERLACED) {
		AGA_build_clist(par, (struct CopList_AGA *)cd->LOFlist,
							 (struct CopList_AGA *)cd->SHFlist, 0);
		AGA_build_clist(par, (struct CopList_AGA *)cd->SHFlist,
							 (struct CopList_AGA *)cd->LOFlist, 1);
	} else
		AGA_build_clist(par, (struct CopList_AGA *)cd->LOFlist, NULL, 0);

	do_vmode = 1;
}


	/*
	 *		Set a single color register. The values supplied are already
	 *		rounded down to the hardware's capabilities (according to the
	 *		entries in the var structure). Return != 0 for invalid regno.
	 */

static int AGA_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
								 u_int transp)
{
	u_long rgb;
	u_short bplcon3 = current_par.hw.AGA.bplcon3;

	if (regno > 255)
		return(1);
	rgb = ((red&0xff)<<16) | ((green&0xff)<<8) | (blue&0xff);
	current_par.hw.AGA.Palette[regno] = rgb;

	/*
	 *		Update the corresponding Hardware Color Register, unless it's Color
	 *		Register 0 and the screen is blanked.
	 *
	 *		The cli()/sti() pair is here to protect bplcon3 from being changed by
	 *		AGA_do_blank()/AGA_do_unblank() during the VBlank.
	 */

	if (regno | !is_blanked) {
		cli();
		custom.bplcon3 = bplcon3 | (regno<<8 & 0xe000);
		custom.color[regno&31] = COLOR_MSB(rgb);
		custom.bplcon3 = bplcon3 | (regno<<8 & 0xe000) | BPC3_LOCT;
		custom.color[regno&31] = COLOR_LSB(rgb);
		custom.bplcon3 = bplcon3;
		sti();
	}

	return(0);
}


	/*
	 *		Read a single color register and split it into
	 *		colors/transparent. Return != 0 for invalid regno.
	 */

static int AGA_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
								 u_int *transp)
{
	u_long rgb;

	if (regno > 255)
		return(1);
	rgb = current_par.hw.AGA.Palette[regno];
	*red = rgb>>16;
	*green = (rgb>>8)&0xff;
	*blue = rgb&0xff;
	return(0);
}


	/*
	 *		Change the video mode (called by VBlank interrupt)
	 */

void AGA_do_vmode(struct display *p)
{
	struct CopDisplay *cd;

	/*
	 *		Switch to the new CopDisplay structure
	 */

	currentcop = 1-currentcop;
	cd = &copdisplay[currentcop];

	/*
	 *		Activate the right Copper Lists
	 */

	custom.cop1lc = (u_short *)cd->Init;
	custom.cop2lc = (u_short *)cd->LOFlist;
	custom.copjmp1 = 0;

	/*
	 *		Make sure we're in a Long Frame (LOF) if the video mode is interlaced.
	 *
	 *		This is always the case if we already were in an interlaced mode, since
	 *		then the VBlank only calls us during a Long Frame.
	 *		But this _is_ necessary if we just switched from a non-interlaced to an
	 *		interlaced mode.
	 */

	if ((current_par.var.vmode & FB_VMODE_MASK) == FB_VMODE_INTERLACED)
		custom.vposw = 0x8000;

	/*
	 *		Update the registers if the video mode changed (i.e. not only pan/wrap)
	 */

	if (full_vmode_change) {
		custom.htotal = current_par.hw.AGA.htotal;
		custom.vtotal = current_par.hw.AGA.vtotal;
		custom.hsstrt = current_par.hw.AGA.hsstrt;
		custom.hsstop = current_par.hw.AGA.hsstop;
		custom.hbstrt = current_par.hw.AGA.hbstrt;
		custom.hbstop = current_par.hw.AGA.hbstop;
		custom.vsstrt = current_par.hw.AGA.vsstrt;
		custom.vsstop = current_par.hw.AGA.vsstop;
		custom.vbstrt = current_par.hw.AGA.vbstrt;
		custom.vbstop = current_par.hw.AGA.vbstop;
		custom.hcenter = current_par.hw.AGA.hcenter;
		custom.beamcon0 = current_par.hw.AGA.beamcon0;
		custom.bplcon1 = current_par.hw.AGA.bplcon1;
		custom.bplcon3 = current_par.hw.AGA.bplcon3;
		custom.bplcon4 = current_par.hw.AGA.bplcon4;
		custom.bpl1mod = current_par.hw.AGA.bpl1mod;
		custom.bpl2mod = current_par.hw.AGA.bpl2mod;

		full_vmode_change = 0;
	}
}


	/*
	 *		Move the cursor (called by VBlank interrupt)
	 */

void AGA_do_movecursor(struct display *p)
{
	long hs, vs, ve;

	if ((current_par.hw.AGA.crsr_x <= -64) ||
		 (current_par.hw.AGA.crsr_x >= current_par.var.xres) ||
		 (current_par.hw.AGA.crsr_y <= -64) ||
		 (current_par.hw.AGA.crsr_y >= current_par.var.yres))
		hs = vs = ve = 0;
	else {
		hs = current_par.hw.AGA.sprstrt_h+current_par.hw.AGA.crsr_x;
		vs = current_par.hw.AGA.sprstrt_v+current_par.hw.AGA.crsr_y;
		ve = vs+64;
	}
	LOFsprite[0] = (vs<<8 & 0xff00) | (hs>>3 & 0x00ff);
	LOFsprite[4] = (ve<<8 & 0xff00) | (vs>>3 & 0x0040) | (ve>>4 & 0x0020) |
						(hs<<3 & 0x0018) | (vs>>6 & 0x0004) | (ve>>7 & 0x0002) |
						(hs>>2 & 0x0001);

	/* TODO: Special case for interlaced or doublescan... */
}


	/*
	 *		Blank the screen (called by VBlank interrupt)
	 */

void AGA_do_blank(struct display *p)
{
	u_short bplcon3 = current_par.hw.AGA.bplcon3;

	custom.dmacon = DMAF_RASTER | DMAF_SPRITE;
	custom.bplcon3 = bplcon3;
	custom.color[0] = 0;
	custom.bplcon3 = bplcon3 | BPC3_LOCT;
	custom.color[0] = 0;
	custom.bplcon3 = bplcon3;
	is_blanked = 1;
}


	/*
	 *		Unblank the screen (called by VBlank interrupt)
	 */

void AGA_do_unblank(struct display *p)
{
	u_long rgb;
	u_short bplcon3 = current_par.hw.AGA.bplcon3;

	if (p->cursor_visible)
		custom.dmacon = DMAF_SETCLR | DMAF_RASTER | DMAF_SPRITE;
	else
		custom.dmacon = DMAF_SETCLR | DMAF_RASTER;

	rgb = current_par.hw.AGA.Palette[0];
	custom.bplcon3 = bplcon3;
	custom.color[0] = COLOR_MSB(rgb);
	custom.bplcon3 = bplcon3 | BPC3_LOCT;
	custom.color[0] = COLOR_LSB(rgb);
	custom.bplcon3 = bplcon3;
	is_blanked = 0;
}


	/*
	 *		Flash the cursor (called by VBlank interrupt)
	 */

void AGA_do_flashcursor(struct display *p)
{
	static int cursorcount = 0;
	static int cursorstate = 0;

	if (p->cursor_visible) {
		if (cursorcount)
			cursorcount--;
		else {
			cursorcount = CRSR_RATE;
			if ((cursorstate = !cursorstate))
				custom.dmacon = DMAF_SETCLR | DMAF_SPRITE;
			else
				custom.dmacon = DMAF_SPRITE;
		}
	} else
		custom.dmacon = DMAF_SPRITE;
}
