
	/*
	 *		Copper Initialisation List
	 */

struct CopInit_ECS {
	copins vsync_hblank;
	copins bplcon0;
	copins diwstrt;
	copins diwstop;
	copins diwhigh;
	copins ddfstrt;
	copins ddfstop;
	copins bplcon2;
	copins sprfix[8];
	copins sprstrtup[16];
	copins wait;
	copins jump;
	copins wait_forever;
};


	/*
	 *		Long Frame/Short Frame Copper List
	 */

struct CopList_ECS {
	copins diwstrt;
	copins diwstop;
	copins diwhigh;
	copins ddfstrt;
	copins ddfstop;
	copins sprpt[2];		/* Sprite 0 */
	copins bplcon0;
	copins bplcon2;
	copins rest[64];
};


	/*
	 *		Hardware Cursor
	 */

struct CursorSprite_ECS {
	u_long Header[1];
	union {
		struct {
			u_long Data[16];
			u_long Trailer[1];
		} NonLaced;
		struct {
			u_long Data[8];
			u_long Trailer[1];
		} Laced;
	} u;
};

struct DummySprite_ECS {
	u_long Header[1];
	u_long Data[1];
	u_long Trailer[1];
};


	/*
	 *		Initialisation
	 *
	 *		Allocate the required chip memory.
	 *		Set the default video mode for this chipset. If a video mode was
	 *		specified on the command line, it will override the default mode.
	 */

static int ECS_init(void)
{
	u_long chipptr;

#ifdef DEBUG
	printf("@@@ ECS_init @@@\n");
#endif

	/*
	 *		Disable Display DMA
	 */

	custom.dmacon = DMAF_ALL | DMAF_MASTER;

	/*
	 *		Set the Default Video Mode
	 */

	if (amifb_mode == -1)
		if (boot_info.bi_un.bi_ami.vblank == 50)
			amifb_mode = PAL_DEFMODE;
		else
			amifb_mode = NTSC_DEFMODE;

	/*
	 *		Allocate Chip RAM Structures
	 */

	if (boot_info.bi_amiga.chip_size > 1048576)
		videomemorysize = VIDEOMEMSIZE_ECS_2M;
	else
		videomemorysize = VIDEOMEMSIZE_ECS_1M;

	chipptr = chipalloc(videomemorysize+
							  2*sizeof(struct CopInit_ECS)+
							  4*sizeof(struct CopList_ECS)+
							  2*sizeof(struct CursorSprite_ECS)+
							  sizeof(struct DummySprite_ECS));

	AssignChunk(videomemory, u_char *, chipptr, videomemorysize);
	AssignChunk(copdisplay[0].Init, copins *, chipptr,
					sizeof(struct CopInit_ECS));
	AssignChunk(copdisplay[0].LOFlist, copins *, chipptr,
					sizeof(struct CopList_ECS));
	AssignChunk(copdisplay[0].SHFlist, copins *, chipptr,
					sizeof(struct CopList_ECS));
	AssignChunk(copdisplay[1].Init, copins *, chipptr,
					sizeof(struct CopInit_ECS));
	AssignChunk(copdisplay[1].LOFlist, copins *, chipptr,
					sizeof(struct CopList_ECS));
	AssignChunk(copdisplay[1].SHFlist, copins *, chipptr,
					sizeof(struct CopList_ECS));
	AssignChunk(LOFsprite, u_long *, chipptr, sizeof(struct CursorSprite_ECS));
	AssignChunk(SHFsprite, u_long *, chipptr, sizeof(struct CursorSprite_ECS));
	AssignChunk(dummysprite, u_long *, chipptr, sizeof(struct DummySprite_ECS));

	/*
	 *		Make sure the Copper has something to do
	 */

#if 0
	ECS_init_copper((struct CopInit_ECS *)copdisplay[0].Init);
#endif

	/*
	 *		Enable Display DMA
	 */

	custom.dmacon = DMAF_SETCLR|DMAF_MASTER|DMAF_RASTER|DMAF_COPPER|DMAF_SPRITE;

	return(0);
}


	/*
	 *		This function should fill in the `fix' structure based on the
	 *		values in the `par' structure.
	 */

static int ECS_encode_fix(struct fb_fix_screeninfo *fix,
								  struct amiga_fb_par *par)
{
#ifdef DEBUG
	printf("@@@ ECS_encode_fix @@@\n");
#endif

	*fix = par->fix;
	return(0);
}


	/*
	 *		Get the video params out of `var'. If a value doesn't fit, round
	 *		it up, if it's too big, return EINVAL.
	 *		Round up in the following order: bits_per_pixel, xres, yres,
	 *		xres_virtual, yres_virtual, xoffset, yoffset, grayscale, bitfields,
	 *		horizontal timing, vertical timing.
	 */

static int ECS_decode_var(struct fb_var_screeninfo *var,
								  struct amiga_fb_par *par)
{
#ifdef DEBUG
	printf("@@@ ECS_decode_var @@@\n");
#endif

	return(-EINVAL);
}


	/*
	 *		Fill the `var' structure based on the values in `par' and maybe
	 *		other values read out of the hardware.
	 */

static int ECS_encode_var(struct fb_var_screeninfo *var,
								  struct amiga_fb_par *par)
{
#ifdef DEBUG
	printf("@@@ ECS_encode_var @@@\n");
#endif

	*var = par->var;
	return(0);
}


	/*
	 *		Fill the hardware's `par' structure.
	 */

static void ECS_get_par(struct amiga_fb_par *par)
{
#ifdef DEBUG
	printf("@@@ ECS_get_par @@@\n");
#endif
}


	/*
	 *		Set the hardware according to `par'.
	 *
	 *		If `pan_only' is TRUE, don't change the whole video mode but only the
	 *		things required for a display pan or ywrap.
	 */

static void ECS_set_par(struct amiga_fb_par *par, u_short pan_only)
{
#ifdef DEBUG
	printf("@@@ ECS_set_par @@@\n");
#endif
}


	/*
	 *		Set a single color register. The values supplied are already
	 *		rounded down to the hardware's capabilities (according to the
	 *		entries in the var structure). Return != 0 for invalid regno.
	 */

static int ECS_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
								 u_int transp)
{
	/* I don't know how ECS color works. Can you fix this, Jes?? */

#ifdef DEBUG
	printf("@@@ ECS_setcolreg @@@\n");
#endif

	return(1);
}


	/*
	 *		Read a single color register and split it into
	 *		colors/transparent. Return != 0 for invalid regno.
	 */

static int ECS_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
								 u_int *transp)
{
	/* I don't know how ECS color works. Can you fix this, Jes?? */

#ifdef DEBUG
	printf("@@@ ECS_getcolreg @@@\n");
#endif

	return(1);
}


	/*
	 *		Change the video mode (called by VBlank interrupt)
	 */

void ECS_do_vmode(struct display *p)
{
}


	/*
	 *		Pan the display (called by VBlank interrupt)
	 */

void ECS_do_pan(struct display *p)
{
}


	/*
	 *		Move the cursor (called by VBlank interrupt)
	 */

void ECS_do_movecursor(struct display *p)
{
}


	/*
	 *		Blank the screen (called by VBlank interrupt)
	 */

void ECS_do_blank(struct display *p)
{
	custom.dmacon = DMAF_RASTER | DMAF_SPRITE;

	/* I don't know how ECS color works. Can you fix this, Jes?? */
}


	/*
	 *		Unblank the screen (called by VBlank interrupt)
	 */

void ECS_do_unblank(struct display *p)
{
	if (p->cursor_visible)
		custom.dmacon = DMAF_SETCLR | DMAF_RASTER | DMAF_SPRITE;
	else
		custom.dmacon = DMAF_SETCLR | DMAF_RASTER;

	/* I don't know how ECS color works. Can you fix this, Jes?? */
}


	/*
	 *		Flash the cursor (called by VBlank interrupt)
	 */

void ECS_do_flashcursor(struct display *p)
{
	static int cursorcount = 0;
	static int cursorstate = 0;

	if (p->cursor_visible) {
		if (cursorcount)
			cursorcount--;
		else {
			cursorcount = CRSR_RATE;
			if ((cursorstate = !cursorstate))
				custom.dmacon = DMAF_SETCLR | DMAF_SPRITE;
			else
				custom.dmacon = DMAF_SPRITE;
		}
	} else
		custom.dmacon = DMAF_SPRITE;
}
