 
/*
 * $Id: chipram.c,v 0.31 1995/06/06 23:49:26 dg2fef Exp $
 * 
 * $Log: chipram.c,v $
 * Revision 0.31  1995/06/06  23:49:26  dg2fef
 * another little cleanup, but seems to be all right now.
 *
 * Revision 0.30  1995/06/05  23:22:13  dg2fef
 * fixed reentrancy problem by disabling interrupts. Another Change to
 * alignment of the memory base in chip_init()
 *
 * Revision 0.21  1995/06/05  18:30:00  dg2fef
 * fixed alignment in chip_init
 *
 * Revision 0.20  1995/06/05  18:23:10  dg2fef
 * cleaned up, fixed a bug in amiga_chip_free, another one in chip_init.
 *
 * Revision 0.11  1995/06/05  16:18:33  dg2fef
 * some typos, and GCC complains for sizeof Chip_Desc :-(
 *
 * Revision 0.1  1995/06/05  16:05:11  dg2fef
 * initial revision
 *
 * 
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/bootinfo.h>
#include <linux/amigahw.h>
#include <asm/system.h>

#define AMIGA_CHIP_MAGIC  0x73BABE88


/* define a 64 Bit integer for alignment */
typedef long long int64;

union chip_desc {
  struct {
    union chip_desc *next;
    unsigned long size;
    unsigned long magic;
  } s;

  int64 x;
};

typedef union chip_desc Chip_Desc;

static Chip_Desc *allocp;
static Chip_Desc *amiga_chip_base;

void amiga_chip_init (void) {

  unsigned long chipsize = boot_info.bi_amiga.chip_size;
  Chip_Desc *p;

  /* not sure about this: Do I have to leave a Gap here for Trap vectors? */
  allocp = amiga_chip_base = p = (Chip_Desc *)(chipaddr+4096);
  chipsize -= 4096;

#ifdef DEBUG
  printk("chip_init: amiga_chip_base at %p\n", (void *)amiga_chip_base);
#endif

  /*
   * dummy entry with size 0, will never be recombined
   * it`s here to speed up freeing memory 
   */

  p->s.magic = AMIGA_CHIP_MAGIC;
  p->s.size = 0;
  p->s.next = p+1;
  p++;

  p->s.magic = AMIGA_CHIP_MAGIC;

  /* note that size is in multiples of the header size */
  p->s.size = chipsize/sizeof(Chip_Desc)-2;
  p->s.next = p-1;

}

void *amiga_chip_alloc (long size) {

  Chip_Desc *p, *q;
  unsigned long nunits;
  unsigned int flags;

  nunits = (size + sizeof(Chip_Desc)-1)/sizeof(Chip_Desc)+1;

  save_flags(flags);
  cli();

  q = allocp;
  for (p=q->s.next; ; q=p, p=p->s.next) {
    if (p->s.size >= nunits) {

      /* found a Chunk, check if it fits */
      if (p->s.size == nunits) {

	/* Chunk fits perfectly */
	q->s.next = p->s.next;
      } else {

	/* Chunk is bigger than needed, so split */
	p->s.size -= nunits;
	p += p->s.size;
	p->s.size = nunits;

	/* fill in MAGIC for debugging */
	p->s.magic = AMIGA_CHIP_MAGIC;
      }
    
      allocp = q;
      restore_flags(flags);
#ifdef DEBUG
      printk("amiga_chip_alloc: alloced %ld units at %p\n", nunits, (void *)(p+1));
#endif
      return (void *)(p+1);
    }

    /* got around the list once, no more memory */
    if (p == allocp) {
      restore_flags(flags);
      return NULL;
    }
  }
}

void amiga_chip_free (void *addr_ptr) {

  Chip_Desc *p, *q;
  unsigned int flags;

  p = (Chip_Desc *)addr_ptr-1;
  if (p->s.magic != AMIGA_CHIP_MAGIC)
    panic("amiga_chip_free: header magic not found");

  q = amiga_chip_base;
  
  save_flags(flags);
  cli();

  while (1) {
    if (q >= q->s.next || p <= q->s.next)
      break;
    q = q->s.next;
  }

  if (p + p->s.size == q->s.next) {
    p->s.size += q->s.next->s.size;
    p->s.next = q->s.next->s.next;
  } else
    p->s.next = q->s.next;

  if (q + q->s.size == p) {
    q->s.size += p->s.size;
    q->s.next = p->s.next;
  } else
    q->s.next = p;

  allocp = q;

  restore_flags(flags);

#ifdef DEBUG
  printk("amiga_chip_free: freed block at %p\n", addr_ptr);
#endif
}
