/*
** linux/amigatypes.h -- Types used in Amiga Linux kernel source
**
** Copyright 1992 by Greg Harp
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
** Created 09/29/92 by Greg Harp
*/

#ifndef _LINUX_AMIGATYPES_H_
#define _LINUX_AMIGATYPES_H_

struct Node {
    struct  Node *ln_Succ;	/* Pointer to next (successor) */
    struct  Node *ln_Pred;	/* Pointer to previous (predecessor) */
    u_char   ln_Type;
    char    ln_Pri;		/* Priority, for sorting */
    char    *ln_Name;		/* ID string, null terminated */
};

/* defined board manufacturers */
#define MANUF_COMMODORE      (0x0202)
#define MANUF_COMMODORE2     (0x0201)	/* for Bridgeboards */
#define MANUF_U_LOWELL       (0x0406)	/* for A2410 */
#define MANUF_AMERISTAR      (0x041D)
#define MANUF_HELFRICH1      (2145)
#define MANUF_HELFRICH2      (2195)
#define MANUF_GVP            (0x07E1)
#define MANUF_VILLAGETRONIC  (0x0877)

#define PROD_A2090A          (0x01)	/* CBM A2090/A2090A HD controller */
#define PROD_A2091           (0x02)	/* CBM A2091 HD controller */
#define PROD_A590            (0x03)	/* CBM A590 HD controller */
#define PROD_A2090B          (0x04)	/* CBM A2090B 2090 Autoboot card */
#define PROD_ARCNET          (0x09)	/* CBM Arcnet card */
#define PROD_CBMRAM          (0x0A)	/* CBM A2052/58.RAM | 590/2091.RAM */
#define PROD_A560RAM         (0x20)	/* CBM A560 memory module */
#define PROD_A2232PROTO      (0x45)	/* CBM A2232 serial prototype */
#define PROD_A2232           (0x46)	/* CBM A2232 serial production */
#define PROD_A2620           (0x50)	/* CBM A2620 68020/RAM card */
#define PROD_A2630           (0x51)	/* CBM A2630 68030/RAM card */
#define PROD_ROMULATOR       (0x60)	/* CBM Romulator card */
#define PROD_A3000TESTFIX    (0x61)	/* CBM A3000 test fixture */
#define PROD_A2065           (0x70)	/* CBM A2065 Ethernet card */
#define PROD_A4091           (0x54)	/* CBM A4091 SCSI HD controller */
#define PROD_A2386SX         (0x67)	/* CBM A2386-SX Bridgeboard */
#define PROD_BRIDGEBOARD     (0x01)	/* CBM Bridgeboard */
#define PROD_A2410           (0x00)	/* CBM A2410 Hires Graphics card */
#define PROD_AMERISTAR2065   (0x01)	/* Ameristar Ethernet card */

#define PROD_GVPIISCSI       (0x0B)
#define PROD_GVPIISCSI_2     (0x09)	/* evidence that the driver works
					   for this product code also */
#define PROD_ARIADNE         (0xC9)	/* Village Tronic Ariadne Ethernet */

#define PROD_RAINBOW3        (33)
#define PROD_PICCOLO_RAM     (5)
#define PROD_PICCOLO_REG     (6)
#define PROD_SD64_RAM        (10)
#define PROD_SD64_REG        (11)

struct ExpansionRom {
    /* -First 16 bytes of the expansion ROM */
    u_char	er_Type;	/* Board type, size and flags */
    u_char	er_Product;	/* Product number, assigned by manufacturer */
    u_char	er_Flags;	/* Flags */
    u_char	er_Reserved03;	/* Must be zero ($ff inverted) */
    u_short	er_Manufacturer; /* Unique ID,ASSIGNED BY COMMODORE-AMIGA! */
    u_long	er_SerialNumber; /* Available for use by manufacturer */
    u_short	er_InitDiagVec; /* Offset to optional "DiagArea" structure */
    u_char	er_Reserved0c;
    u_char	er_Reserved0d;
    u_char	er_Reserved0e;
    u_char	er_Reserved0f;
};

struct ConfigDev {
    struct Node 	cd_Node;
    u_char		cd_Flags;	/* (read/write) */
    u_char		cd_Pad; 	/* reserved */
    struct ExpansionRom cd_Rom; 	/* copy of board's expansion ROM */
    void		*cd_BoardAddr;	/* where in memory the board was placed */
    u_long		cd_BoardSize;	/* size of board in bytes */
    u_short		cd_SlotAddr;	/* which slot number (PRIVATE) */
    u_short		cd_SlotSize;	/* number of slots (PRIVATE) */
    void		*cd_Driver;	/* pointer to node of driver */
    struct ConfigDev	*cd_NextCD;	/* linked list of drivers to config */
    u_long		cd_Unused[4];	/* for whatever the driver wants */
};

/*
 * Different models of Amiga
 */
#define AMI_UNKNOWN  (0)
#define AMI_500      (1)
#define AMI_2000     (2)
#define AMI_3000     (3)
#define AMI_4000     (4)
#define AMI_1200     (5)

/*
 * chipsets
 */
#define CS_STONEAGE (0)
#define CS_OCS	    (1)
#define CS_ECS	    (2)
#define CS_AGA	    (3)

/*
 * number of CIA Eclock ticks for a single scheduling tick (jiffy)
 */

#define NTSC_JIFFY_TICKS   (7159)
#define PAL_JIFFY_TICKS    (7094)

#endif /* linux/amigatypes.h */
